###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
import itertools
from pyomo.environ import (
    ConcreteModel,
    value,
    Param,
    Var,
    Constraint,
    Expression,
    Objective,
    TransformationFactory,
    NonNegativeReals,
    RangeSet,
    check_optimal_termination,
    assert_optimal_termination,
    units as pyunits,
)
from pyomo.network import Arc
from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import (
    propagate_state as _pro_state,
)
from idaes.models.unit_models import Mixer, Separator, Product, Feed
from idaes.core import UnitModelCostingBlock
import idaes.core.util.scaling as iscale
from idaes.core.util.misc import StrEnum

import watertap.property_models.NaCl_prop_pack as props
from watertap.core.util.initialization import (
    assert_degrees_of_freedom,
)
from watertap.unit_models.reverse_osmosis_0D import (
    ReverseOsmosis0D,
    ConcentrationPolarizationType,
)
from watertap.unit_models.osmotically_assisted_reverse_osmosis_0D import (
    OsmoticallyAssistedReverseOsmosis0D,
    MassTransferCoefficient,
    PressureChangeType,
)
from watertap.unit_models.pressure_changer import Pump, EnergyRecoveryDevice
from watertap.costing import WaterTAPCosting
from watertap.costing.unit_models.pump import cost_low_pressure_pump


class ERDtype(StrEnum):
    pump_as_turbine = "pump_as_turbine"


def erd_type_not_found(erd_type):
    raise NotImplementedError(
        "erd_type was {}, but can only " "be pump_as_turbine" "".format(erd_type.value)
    )


def propagate_state(arc):
    _pro_state(arc)
    print(arc.destination.name)
    arc.destination.display()


def main(number_of_stages, system_recovery, erd_type=ERDtype.pump_as_turbine):
    # set up solver
    solver = get_solver()

    # build, set, and initialize
    m = build(number_of_stages=number_of_stages, erd_type=erd_type)
    set_operating_conditions(m)
    initialize_system(
        m,
        number_of_stages,
        solvent_multiplier=0.5,
        solute_multiplier=0.7,
        solver=solver,
    )

    optimize_set_up(
        m, number_of_stages=number_of_stages, water_recovery=system_recovery
    )

    results = solve(m, solver=solver)
    assert_optimal_termination(results)

    print("\n***---Optimization results---***")
    display_system(m)
    display_design(m)
    if erd_type == ERDtype.pump_as_turbine:
        display_state(m)
    else:
        pass

    return m


def build(number_of_stages, erd_type=ERDtype.pump_as_turbine):

    # flowsheet set up
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.erd_type = erd_type
    m.fs.properties = props.NaClParameterBlock()
    m.fs.costing = WaterTAPCosting()

    # stage set up
    m.fs.NumberOfStages = Param(initialize=number_of_stages)
    m.fs.Stages = RangeSet(m.fs.NumberOfStages)
    m.fs.NonFirstStages = RangeSet(2, m.fs.NumberOfStages)
    m.fs.NonFinalStages = RangeSet(m.fs.NumberOfStages - 1)
    if number_of_stages > 2:
        m.fs.IntermediateStages = RangeSet(2, m.fs.NumberOfStages - 1)
    else:
        m.fs.IntermediateStages = RangeSet(0)
    m.fs.FirstStage = m.fs.Stages.first()
    m.fs.LastStage = m.fs.Stages.last()

    # Control volume flow blocks
    m.fs.feed = Feed(property_package=m.fs.properties)
    m.fs.product = Product(property_package=m.fs.properties)
    m.fs.disposal = Product(property_package=m.fs.properties)

    # --- Main pump ---
    m.fs.PrimaryPumps = Pump(m.fs.Stages, property_package=m.fs.properties)
    for pump in m.fs.PrimaryPumps.values():
        pump.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
        )

    # --- Recycle pump ---
    m.fs.RecyclePumps = Pump(m.fs.NonFirstStages, property_package=m.fs.properties)
    for pump in m.fs.RecyclePumps.values():
        pump.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing, costing_method=cost_low_pressure_pump
        )

    m.fs.total_pump_work = Expression(
        expr=sum(
            pyunits.convert(pump.work_mechanical[0], to_units=pyunits.kW)
            for pump in itertools.chain(
                m.fs.PrimaryPumps.values(), m.fs.RecyclePumps.values()
            )
        )
    )

    # --- Reverse Osmosis Block ---
    m.fs.RO = ReverseOsmosis0D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        pressure_change_type=PressureChangeType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        has_full_reporting=True,
    )
    m.fs.RO.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    # --- Osmotically Assisted Reverse Osmosis Block ---
    m.fs.OAROUnits = OsmoticallyAssistedReverseOsmosis0D(
        m.fs.NonFinalStages,
        property_package=m.fs.properties,
        has_pressure_change=True,
        pressure_change_type=PressureChangeType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        has_full_reporting=True,
    )
    for stage in m.fs.OAROUnits.values():
        stage.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method_arguments={"oaro_type": "high_pressure"},
        )

    # --- ERD blocks ---
    if erd_type == ERDtype.pump_as_turbine:
        # add energy recovery turbine block
        m.fs.EnergyRecoveryDevices = EnergyRecoveryDevice(
            m.fs.Stages, property_package=m.fs.properties
        )
        # add costing for ERD config
        for erd in m.fs.EnergyRecoveryDevices.values():
            erd.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    else:
        erd_type_not_found(erd_type)

    # --- Separators ---
    m.fs.Separators = Separator(
        m.fs.NonFirstStages,
        property_package=m.fs.properties,
        outlet_list=["treat", "purge"],
    )
    if number_of_stages > 1:
        separator_list = ["product"]
        for i in range(2, number_of_stages + 1):
            separator_list.append("makeup" + str(i))
        m.fs.ProductSeparator = Separator(
            property_package=m.fs.properties, outlet_list=separator_list
        )

    # --- Mixers ---
    m.fs.Mixers = Mixer(
        m.fs.NonFirstStages,
        property_package=m.fs.properties,
        inlet_list=["treat", "makeup"],
    )
    if number_of_stages > 1:
        mixer_list = []
        for i in range(1, number_of_stages + 1):
            mixer_list.append("purge" + str(i))
        m.fs.WasteMixer = Mixer(property_package=m.fs.properties, inlet_list=mixer_list)

    m.fs.recovered_pump_work = Expression(
        expr=sum(
            pyunits.convert(erd.work_mechanical[0], to_units=pyunits.kW)
            for erd in m.fs.EnergyRecoveryDevices.values()
        )
    )
    m.fs.net_pump_work = Expression(
        expr=m.fs.total_pump_work + m.fs.recovered_pump_work
    )

    # additional parameters, variables or expressions
    m.fs.oaro_min_pressure = Param(initialize=10e5, units=pyunits.Pa, mutable=True)
    m.fs.oaro_max_pressure = Param(initialize=65e5, units=pyunits.Pa, mutable=True)
    m.fs.ro_min_pressure = Param(initialize=10e5, units=pyunits.Pa, mutable=True)
    m.fs.ro_max_pressure = Param(initialize=85e5, units=pyunits.Pa, mutable=True)
    m.fs.recycle_pump_min_pressure = Param(
        initialize=1e5, units=pyunits.Pa, mutable=True
    )
    m.fs.recycle_pump_max_pressure = Param(
        initialize=85e5, units=pyunits.Pa, mutable=True
    )

    # process costing and add system level metrics
    m.fs.costing.utilization_factor.fix(0.9)
    m.fs.costing.TIC.fix(2)
    m.fs.costing.factor_maintenance_labor_chemical.fix(0.03)
    m.fs.costing.factor_capital_annualization.fix(0.1)
    m.fs.costing.electricity_cost.set_value(0.07)
    m.fs.costing.reverse_osmosis.factor_membrane_replacement.fix(0.15)
    m.fs.costing.reverse_osmosis.membrane_cost.fix(30)
    m.fs.costing.reverse_osmosis.high_pressure_membrane_cost.fix(50)
    m.fs.costing.high_pressure_pump.cost.fix(53 / 1e5 * 3600)
    m.fs.costing.energy_recovery_device.pressure_exchanger_cost.fix(535)

    m.fs.costing.cost_process()

    product_flow_vol_total = m.fs.product.properties[0].flow_vol
    m.fs.costing.add_annual_water_production(product_flow_vol_total)
    m.fs.costing.add_LCOW(product_flow_vol_total)
    m.fs.costing.add_specific_energy_consumption(product_flow_vol_total)
    m.fs.costing.add_specific_electrical_carbon_intensity(product_flow_vol_total)

    # Expressions for parameter sweep -----------------------------------------
    # Final permeate concentration as mass fraction
    m.fs.product.properties[0].mass_frac_phase_comp

    # Touch feed concentration as mass concentration
    m.fs.feed.properties[0].conc_mass_phase_comp

    # Touch final brine concentration as mass concentration
    m.fs.disposal.properties[0].conc_mass_phase_comp

    # Touch final brine concentration as mass fraction
    m.fs.disposal.properties[0].mass_frac_phase_comp

    m.fs.system_salt_rejection = Expression(
        expr=1
        - m.fs.product.properties[0].conc_mass_phase_comp["Liq", "NaCl"]
        / m.fs.feed.properties[0].conc_mass_phase_comp["Liq", "NaCl"]
    )
    m.fs.annual_feed = Expression(
        expr=pyunits.convert(
            m.fs.feed.properties[0].flow_vol, to_units=pyunits.m**3 / pyunits.year
        )
        * m.fs.costing.utilization_factor
    )
    m.fs.final_permeate_concentration = Expression(
        expr=m.fs.product.flow_mass_phase_comp[0, "Liq", "NaCl"]
        / sum(m.fs.product.flow_mass_phase_comp[0, "Liq", j] for j in ["H2O", "NaCl"])
    )
    m.fs.costing.add_LCOW(m.fs.feed.properties[0].flow_vol, name="LCOW_feed")

    m.fs.costing.add_specific_energy_consumption(
        m.fs.feed.properties[0].flow_vol, name="specific_energy_consumption_feed"
    )

    m.fs.costing.primary_pump_capex_lcow = Expression(
        expr=m.fs.costing.factor_capital_annualization
        * sum(m.fs.PrimaryPumps[n].costing.capital_cost for n in m.fs.Stages)
        / m.fs.costing.annual_water_production
    )

    m.fs.total_membrane_area = Expression(
        expr=sum(oaro.area for oaro in m.fs.OAROUnits.values()) + m.fs.RO.area
    )

    m.fs.costing.recycle_pump_capex_lcow = Expression(
        expr=m.fs.costing.factor_capital_annualization
        * (
            sum(m.fs.RecyclePumps[n].costing.capital_cost for n in m.fs.NonFirstStages)
            if number_of_stages > 1
            else 0 * m.fs.costing.base_currency
        )
        / m.fs.costing.annual_water_production,
    )

    m.fs.costing.erd_capex_lcow = Expression(
        expr=m.fs.costing.factor_capital_annualization
        * sum(erd.costing.capital_cost for erd in m.fs.EnergyRecoveryDevices.values())
        / m.fs.costing.annual_water_production
    )

    m.fs.costing.electricity_lcow = Expression(
        expr=m.fs.costing.aggregate_flow_costs["electricity"]
        * m.fs.costing.utilization_factor
        / m.fs.costing.annual_water_production
    )

    m.fs.costing.pumping_energy_aggregate_lcow = Expression(
        expr=m.fs.costing.factor_total_investment
        * (
            m.fs.costing.primary_pump_capex_lcow
            + (
                m.fs.costing.recycle_pump_capex_lcow
                if number_of_stages > 1
                else 0 * m.fs.costing.base_currency
            )
            + m.fs.costing.erd_capex_lcow
        )
        * (
            1
            + m.fs.costing.factor_maintenance_labor_chemical
            / m.fs.costing.factor_total_investment
            / m.fs.costing.factor_capital_annualization
        )
        + m.fs.costing.electricity_lcow
    )

    m.fs.costing.membrane_capex_lcow = Expression(
        expr=m.fs.costing.factor_capital_annualization
        * (
            sum(m.fs.OAROUnits[n].costing.capital_cost for n in m.fs.NonFinalStages)
            + m.fs.RO.costing.capital_cost
        )
        / m.fs.costing.annual_water_production
    )

    m.fs.costing.indirect_capex_lcow = Expression(
        expr=m.fs.costing.factor_capital_annualization
        * (m.fs.costing.total_capital_cost - m.fs.costing.aggregate_capital_cost)
        / m.fs.costing.annual_water_production
    )

    m.fs.costing.membrane_replacement_lcow = Expression(
        expr=(
            sum(
                m.fs.OAROUnits[n].costing.fixed_operating_cost
                for n in m.fs.NonFinalStages
            )
            + m.fs.RO.costing.fixed_operating_cost
        )
        / m.fs.costing.annual_water_production
    )

    m.fs.costing.chemical_labor_maintenance_lcow = Expression(
        expr=m.fs.costing.maintenance_labor_chemical_operating_cost
        / m.fs.costing.annual_water_production
    )

    m.fs.costing.membrane_aggregate_lcow = Expression(
        expr=m.fs.costing.factor_total_investment
        * m.fs.costing.membrane_capex_lcow
        * (
            1
            + m.fs.costing.factor_maintenance_labor_chemical
            / m.fs.costing.factor_total_investment
            / m.fs.costing.factor_capital_annualization
        )
        + m.fs.costing.membrane_replacement_lcow
    )

    # system water recovery
    m.fs.water_recovery = Var(
        initialize=0.5,
        bounds=(0, 1),
        domain=NonNegativeReals,
        units=pyunits.dimensionless,
        doc="System Volumetric Recovery of Water",
    )
    m.fs.eq_water_recovery = Constraint(
        expr=(
            m.fs.feed.properties[0].flow_vol_phase["Liq"]
            + sum(
                m.fs.Mixers[stage].makeup_state[0].flow_vol_phase["Liq"]
                for stage in m.fs.NonFirstStages
            )
        )
        * m.fs.water_recovery
        == m.fs.product.properties[0].flow_vol_phase["Liq"]
    )

    m.fs.mass_water_recovery = Var(
        initialize=0.5,
        bounds=(0, 1),
        domain=NonNegativeReals,
        units=pyunits.dimensionless,
        doc="System Mass Recovery of Water",
    )
    m.fs.eq_mass_water_recovery = Constraint(
        expr=(
            m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"]
            + sum(
                m.fs.Mixers[stage].makeup_state[0].flow_mass_phase_comp["Liq", "H2O"]
                for stage in m.fs.NonFirstStages
            )
        )
        * m.fs.mass_water_recovery
        == m.fs.product.properties[0].flow_mass_phase_comp["Liq", "H2O"]
    )

    # connections
    if erd_type == ERDtype.pump_as_turbine:
        last_stage = m.fs.LastStage
        first_stage = m.fs.FirstStage

        # Connect the feed to the first pump
        m.fs.feed_to_pump = Arc(
            source=m.fs.feed.outlet, destination=m.fs.PrimaryPumps[first_stage].inlet
        )

        if number_of_stages > 1:
            # Connect the Pump n to OARO n
            m.fs.pump_to_OARO = Arc(
                m.fs.NonFinalStages,
                rule=lambda fs, n: {
                    "source": fs.PrimaryPumps[n].outlet,
                    "destination": fs.OAROUnits[n].feed_inlet,
                },
            )

            # Connect OARO n permeate outlet to Pump n+1
            m.fs.OARO_to_pump = Arc(
                m.fs.NonFinalStages,
                rule=lambda fs, n: {
                    "source": fs.OAROUnits[n].permeate_outlet,
                    "destination": fs.PrimaryPumps[n + 1].inlet,
                },
            )

            # Connect OARO n feed outlet to EnergyRecoveryDevice n
            m.fs.OARO_to_ERD = Arc(
                m.fs.NonFinalStages,
                rule=lambda fs, n: {
                    "source": fs.OAROUnits[n].feed_outlet,
                    "destination": fs.EnergyRecoveryDevices[n].inlet,
                },
            )

            # Connect EnergyRecoveryDevice n to Separator n
            m.fs.ERD_to_separator = Arc(
                m.fs.NonFirstStages,
                rule=lambda fs, n: {
                    "source": fs.EnergyRecoveryDevices[n].outlet,
                    "destination": fs.Separators[n].inlet,
                },
            )

            m.fs.separator_to_mixer = Arc(
                m.fs.NonFirstStages,
                rule=lambda fs, n: {
                    "source": fs.Separators[n].treat,
                    "destination": fs.Mixers[n].treat,
                },
            )

            m.fs.makeup_to_mixer = Arc(
                m.fs.NonFirstStages,
                rule=lambda fs, n: {
                    "source": getattr(fs.ProductSeparator, "makeup" + str(n)),
                    "destination": fs.Mixers[n].makeup,
                },
            )

            m.fs.mixer_to_recyclepump = Arc(
                m.fs.NonFirstStages,
                rule=lambda fs, n: {
                    "source": fs.Mixers[n].outlet,
                    "destination": fs.RecyclePumps[n].inlet,
                },
            )

            # Connect RecyclePumps n to OARO n-1 permeate_inlet
            m.fs.recyclepump_to_OARO = Arc(
                m.fs.NonFirstStages,
                rule=lambda fs, n: {
                    "source": fs.RecyclePumps[n].outlet,
                    "destination": fs.OAROUnits[n - 1].permeate_inlet,
                },
            )

            # Connect purge streams to waste mixer
            m.fs.purge_to_wastemixer = Arc(
                m.fs.NonFirstStages,
                rule=lambda fs, n: {
                    "source": fs.Separators[n].purge,
                    "destination": getattr(fs.WasteMixer, "purge" + str(n)),
                },
            )

        if number_of_stages > 1:
            # Connect first EnergyRecoveryDevice to Mixer
            m.fs.ERD_to_wastemixer = Arc(
                source=m.fs.EnergyRecoveryDevices[first_stage].outlet,
                destination=m.fs.WasteMixer.purge1,
            )

            # Connect waste mixer to disposal
            m.fs.wastemixer_to_disposal = Arc(
                source=m.fs.WasteMixer.outlet,
                destination=m.fs.disposal.inlet,
            )
        else:
            m.fs.ERD_to_disposal = Arc(
                source=m.fs.EnergyRecoveryDevices[first_stage].outlet,
                destination=m.fs.disposal.inlet,
            )

        # Connect the last Pump to RO
        m.fs.pump_to_ro = Arc(
            source=m.fs.PrimaryPumps[last_stage].outlet, destination=m.fs.RO.inlet
        )

        # Connect the primary RO permeate to the product
        # m.fs.ro_to_product = Arc(
        #     source=m.fs.RO.permeate, destination=m.fs.product.inlet
        # )
        # Connect the primary RO permeate to the product
        if number_of_stages > 1:
            m.fs.ro_to_productseparator = Arc(
                source=m.fs.RO.permeate, destination=m.fs.ProductSeparator.inlet
            )
            m.fs.productseparator_to_product = Arc(
                source=m.fs.ProductSeparator.product, destination=m.fs.product.inlet
            )
        else:
            m.fs.ro_to_product = Arc(
                source=m.fs.RO.permeate, destination=m.fs.product.inlet
            )

        # Connect RO retentate to the last EnergyRecoveryDevice
        m.fs.ro_to_ERD = Arc(
            source=m.fs.RO.retentate,
            destination=m.fs.EnergyRecoveryDevices[last_stage].inlet,
        )

    else:
        # this case should be caught in the previous conditional
        erd_type_not_found(erd_type)

    TransformationFactory("network.expand_arcs").apply_to(m)

    # scaling
    # set default property values
    m.fs.properties.set_default_scaling("flow_mass_phase_comp", 1, index=("Liq", "H2O"))
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e2, index=("Liq", "NaCl")
    )
    # set unit model values
    for pump in m.fs.PrimaryPumps.values():
        iscale.set_scaling_factor(pump.control_volume.work, 1e-3)

    for pump in m.fs.RecyclePumps.values():
        iscale.set_scaling_factor(pump.control_volume.work, 1e-3)

    iscale.set_scaling_factor(m.fs.RO.area, 1e-2)

    for stage in m.fs.NonFinalStages:
        iscale.set_scaling_factor(m.fs.OAROUnits[stage].area, 1e-2)

    m.fs.feed.properties[0].flow_vol_phase["Liq"]
    m.fs.feed.properties[0].mass_frac_phase_comp["Liq", "NaCl"]

    if erd_type == ERDtype.pump_as_turbine:
        for erd in m.fs.EnergyRecoveryDevices.values():
            iscale.set_scaling_factor(erd.control_volume.work, 1e-3)
    else:
        erd_type_not_found(erd_type)
    # unused scaling factors needed by IDAES base costing module
    # calculate and propagate scaling factors
    iscale.calculate_scaling_factors(m)

    return m


def set_operating_conditions(
    m,
    solver=None,
):

    if solver is None:
        solver = get_solver()

    # ---specifications---
    # feed
    # state variables
    pressure_atmospheric = 101325
    feed_temperature = 273.15 + 25
    m.fs.feed.properties[0].pressure.fix(pressure_atmospheric)  # feed pressure [Pa]
    m.fs.feed.properties[0].temperature.fix(feed_temperature)  # feed temperature [K]

    Cin = 75000e-6
    Qin = 5.416667e-3
    m.fs.feed.properties.calculate_state(
        var_args={
            ("mass_frac_phase_comp", ("Liq", "NaCl")): value(
                Cin
            ),  # feed mass concentration
            ("flow_vol_phase", "Liq"): Qin,
        },  # volumetric feed flowrate [-]
        hold_state=True,  # fixes the calculated component mass flow rates
    )

    # primary pumps
    for idx, pump in m.fs.PrimaryPumps.items():
        # pump.control_volume.properties_out[0].pressure = 10e5 + 65e5 / float(idx)
        if idx == m.fs.Stages.last():
            pump.control_volume.properties_out[0].pressure = 85e5
        else:
            pump.control_volume.properties_out[0].pressure = 65e5
        pump.control_volume.properties_out[0].pressure.fix()
        pump.efficiency_pump.fix(0.75)

    for idx, pump in m.fs.RecyclePumps.items():
        pump.control_volume.properties_out[0].pressure = 2e5
        pump.control_volume.properties_out[0].pressure.fix()
        pump.efficiency_pump.fix(0.75)

    # Initialize OARO
    A_OARO = 1.0e-12
    B_OARO = 8.0e-8
    spacer_porosity = 0.75

    for idx, stage in m.fs.OAROUnits.items():
        stage.A_comp.fix(A_OARO)
        stage.B_comp.fix(B_OARO)

        stage.structural_parameter.fix(1200e-6)

        stage.permeate_side.channel_height.fix(2e-3)
        stage.permeate_side.spacer_porosity.fix(spacer_porosity)
        stage.feed_side.channel_height.fix(2e-3)
        stage.feed_side.spacer_porosity.fix(spacer_porosity)
        stage.feed_side.velocity[0, 0].fix(0.1)
        stage.recovery_mass_phase_comp[0, "Liq", "H2O"].fix(0.3)

    # RO unit
    A_RO = 4.2e-12
    B_RO = 3.5e-8
    width = 5 * Qin / 1e-3  # effective membrane width [m]
    m.fs.RO.A_comp.fix(A_RO)  # membrane water permeability coefficient [m/s-Pa]
    m.fs.RO.B_comp.fix(B_RO)  # membrane salt permeability coefficient [m/s]
    m.fs.RO.feed_side.channel_height.fix(2e-3)  # channel height in membrane stage [m]
    m.fs.RO.feed_side.spacer_porosity.fix(
        spacer_porosity
    )  # spacer porosity in membrane stage [-]
    m.fs.RO.permeate.pressure[0].fix(101325)  # atmospheric pressure [Pa]
    m.fs.RO.width.fix(width / float(m.fs.NumberOfStages))  # stage width [m]
    m.fs.RO.recovery_mass_phase_comp[0, "Liq", "H2O"].fix(0.3)

    # ERD unit
    if m.fs.erd_type == ERDtype.pump_as_turbine:
        # energy recovery turbine - efficiency and outlet pressure
        for erd in m.fs.EnergyRecoveryDevices.values():
            erd.efficiency_pump.fix(0.9)
            erd.control_volume.properties_out[0].pressure.fix(pressure_atmospheric)
    else:
        erd_type_not_found(m.fs.erd_type)

    # Makeup
    for i in range(2, m.fs.NumberOfStages + 1):
        m.fs.ProductSeparator.split_fraction[:, "makeup" + str(i)].fix(0)

    # Separator
    for sep in m.fs.Separators.values():
        sep.split_fraction[:, "treat"] = 1
        sep.split_fraction[:, "treat"].fix()

    # check degrees of freedom
    if degrees_of_freedom(m) != 0:
        raise RuntimeError(
            "The set_operating_conditions function resulted in {} "
            "degrees of freedom rather than 0. This error suggests "
            "that too many or not enough variables are fixed for a "
            "simulation.".format(degrees_of_freedom(m))
        )


def purge_initializer(sep, oaro, solute_multiplier):
    split_fraction_treat = value(
        oaro.feed_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"]
        * solute_multiplier
        / sep.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"]
    )
    if split_fraction_treat <= 1:
        sep.split_fraction[:, "treat"].value = (
            oaro.feed_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"]
            * solute_multiplier
            / sep.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"]
        )
    else:
        sep.split_fraction[:, "treat"].value = 1


def makeup_initializer(mixer, oaro, solvent_multiplier):
    makeup_flow_mass = value(
        oaro.feed_inlet.flow_mass_phase_comp[0, "Liq", "H2O"] * solvent_multiplier
        - mixer.treat.flow_mass_phase_comp[0, "Liq", "H2O"]
    )
    if makeup_flow_mass >= 0:
        mixer.makeup_state[0].flow_mass_phase_comp["Liq", "H2O"].value = (
            oaro.feed_inlet.flow_mass_phase_comp[0, "Liq", "H2O"] * solvent_multiplier
            - mixer.treat.flow_mass_phase_comp[0, "Liq", "H2O"]
        )
    else:
        mixer.makeup_state[0].flow_mass_phase_comp["Liq", "H2O"].value = 1e-9
    mixer.makeup_state[0].flow_mass_phase_comp["Liq", "NaCl"].value = 0


def recycle_pump_initializer(pump, oaro, solvent_multiplier, solute_multiplier):

    feed_temperature = 273.15 + 25
    pump.control_volume.properties_out[0].temperature.value = feed_temperature
    pump.control_volume.properties_out[0].flow_mass_phase_comp["Liq", "H2O"].value = (
        oaro.feed_inlet.flow_mass_phase_comp[0, "Liq", "H2O"] * solvent_multiplier
    )
    pump.control_volume.properties_out[0].flow_mass_phase_comp["Liq", "NaCl"].value = (
        oaro.feed_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"] * solute_multiplier
    )


def solve(model, solver=None, tee=False, raise_on_failure=False):
    # ---solving---
    if solver is None:
        solver = get_solver(options={"bound_push": 1e-2})

    results = solver.solve(model, tee=tee)
    if check_optimal_termination(results):
        return results
    msg = (
        "The current configuration is infeasible. Please adjust the decision variables."
    )
    if raise_on_failure:
        raise RuntimeError(msg)
    else:
        print(msg)
        return results


def initialize_loop(m, solvent_multiplier, solute_multiplier):

    for stage in m.fs.IntermediateStages:
        propagate_state(m.fs.OARO_to_pump[stage - 1])
        m.fs.PrimaryPumps[stage].initialize()

        # ---initialize loop---
        propagate_state(m.fs.pump_to_OARO[stage])
        recycle_pump_initializer(
            m.fs.RecyclePumps[stage + 1],
            m.fs.OAROUnits[stage],
            solvent_multiplier=solvent_multiplier,
            solute_multiplier=solute_multiplier,
        )
        propagate_state(m.fs.recyclepump_to_OARO[stage + 1])
        m.fs.OAROUnits[stage].initialize()

        propagate_state(m.fs.OARO_to_ERD[stage])
        m.fs.EnergyRecoveryDevices[stage].initialize()
        propagate_state(m.fs.ERD_to_separator[stage])
        purge_initializer(
            m.fs.Separators[stage],
            m.fs.OAROUnits[stage - 1],
            solute_multiplier=solute_multiplier,
        )
        m.fs.Separators[stage].initialize()

        propagate_state(m.fs.separator_to_mixer[stage])
        makeup_initializer(
            m.fs.Mixers[stage],
            m.fs.OAROUnits[stage - 1],
            solvent_multiplier=solvent_multiplier,
        )
        m.fs.Mixers[stage].initialize()
        propagate_state(m.fs.mixer_to_recyclepump[stage])
        m.fs.RecyclePumps[stage].initialize()

        propagate_state(m.fs.recyclepump_to_OARO[stage])
        propagate_state(m.fs.pump_to_OARO[stage - 1])
        m.fs.OAROUnits[stage - 1].initialize()


def initialize_system(
    m,
    number_of_stages=None,
    solvent_multiplier=None,
    solute_multiplier=None,
    solver=None,
):
    if solver is None:
        solver = get_solver()

    last_stage = m.fs.LastStage
    first_stage = m.fs.FirstStage

    # ---initialize feed block---
    m.fs.feed.initialize()

    propagate_state(m.fs.feed_to_pump)
    m.fs.PrimaryPumps[first_stage].initialize()

    # ---initialize first OARO unit---
    if number_of_stages > 1:
        propagate_state(m.fs.pump_to_OARO[first_stage])
        recycle_pump_initializer(
            m.fs.RecyclePumps[first_stage + 1],
            m.fs.OAROUnits[first_stage],
            solvent_multiplier=solvent_multiplier,
            solute_multiplier=solute_multiplier,
        )
        propagate_state(m.fs.recyclepump_to_OARO[first_stage + 1])
        m.fs.OAROUnits[first_stage].initialize()

        if number_of_stages > 2:
            initialize_loop(m, solvent_multiplier, solute_multiplier)

        # ---initialize RO loop---
        propagate_state(m.fs.OARO_to_pump[last_stage - 1])
        m.fs.PrimaryPumps[last_stage].initialize()

    propagate_state(m.fs.pump_to_ro)
    m.fs.RO.initialize()

    propagate_state(m.fs.ro_to_ERD)
    m.fs.EnergyRecoveryDevices[last_stage].initialize()

    if number_of_stages > 1:
        purge_initializer(
            m.fs.Separators[last_stage],
            m.fs.OAROUnits[last_stage - 1],
            solute_multiplier=solute_multiplier,
        )
        propagate_state(m.fs.ERD_to_separator[last_stage])
        m.fs.Separators[last_stage].initialize()
        propagate_state(m.fs.separator_to_mixer[last_stage])
        makeup_initializer(
            m.fs.Mixers[last_stage],
            m.fs.OAROUnits[last_stage - 1],
            solvent_multiplier=solvent_multiplier,
        )
        m.fs.Mixers[last_stage].initialize()
        propagate_state(m.fs.mixer_to_recyclepump[last_stage])
        m.fs.RecyclePumps[last_stage].initialize()

        propagate_state(m.fs.recyclepump_to_OARO[last_stage])
        propagate_state(m.fs.pump_to_OARO[last_stage - 1])
        m.fs.OAROUnits[last_stage - 1].initialize()

        # ---initialize first ERD---
        propagate_state(m.fs.OARO_to_ERD[first_stage])
        m.fs.EnergyRecoveryDevices[first_stage].initialize()

    m.fs.costing.initialize()


def optimize_set_up(
    m,
    number_of_stages=None,
    water_recovery=None,
):
    # add objective
    m.fs.objective = Objective(expr=m.fs.costing.LCOW)

    # Primary pumps
    for idx, pump in m.fs.PrimaryPumps.items():
        pump.control_volume.properties_out[0].pressure.unfix()
        pump.deltaP.setlb(0)
        if idx == m.fs.Stages.last():
            pump.max_ro_pressure_con = Constraint(
                expr=(
                    m.fs.ro_min_pressure,
                    pump.control_volume.properties_out[0].pressure,
                    m.fs.ro_max_pressure,
                )
            )
        else:
            pump.max_oaro_pressure_con = Constraint(
                expr=(
                    m.fs.oaro_min_pressure,
                    pump.control_volume.properties_out[0].pressure,
                    m.fs.oaro_max_pressure,
                )
            )

    # Recycle pumps
    for idx, pump in m.fs.RecyclePumps.items():
        pump.control_volume.properties_out[0].pressure.unfix()
        pump.max_recycle_pump_pressure_con = Constraint(
            expr=(
                m.fs.recycle_pump_min_pressure,
                pump.control_volume.properties_out[0].pressure,
                m.fs.recycle_pump_max_pressure,
            )
        )
        iscale.constraint_scaling_transform(
            pump.max_recycle_pump_pressure_con,
            iscale.get_scaling_factor(pump.control_volume.properties_out[0].pressure),
        )
        pump.deltaP.setlb(0)

    # OARO Units
    for stage in m.fs.OAROUnits.values():
        stage.recovery_mass_phase_comp[0, "Liq", "H2O"].unfix()

        stage.area.unfix()
        stage.area.setlb(1)
        stage.area.setub(20000)

        stage.width.unfix()
        stage.width.setlb(0.1)
        stage.width.setub(1000)

        stage.feed_side.velocity[0, 0].unfix()
        stage.feed_side.velocity[0, 0].setlb(0)
        stage.feed_side.velocity[0, 0].setub(1)

        stage.feed_side.N_Re[0, 0].setlb(100)
        stage.feed_side.N_Re[0, 0].setub(2000)

        stage.feed_side.N_Re[0, 1].setlb(100)
        stage.feed_side.N_Re[0, 1].setub(2000)

        stage.permeate_side.N_Re[0, 0].setlb(100)
        stage.permeate_side.N_Re[0, 0].setub(2000)

        stage.permeate_side.N_Re[0, 1].setlb(100)
        stage.permeate_side.N_Re[0, 1].setub(2000)

        stage.permeate_outlet.pressure[0].unfix()
        stage.permeate_outlet.pressure[0].setlb(101325)

        stage.oaro_avg_water_flux_con = Constraint(
            expr=(
                0.1,
                pyunits.convert(
                    stage.flux_mass_phase_comp_avg[0, "Liq", "H2O"]
                    / stage.dens_solvent,
                    to_units=pyunits.L / pyunits.m**2 / pyunits.hr,
                ),
                10,
            )
        )

        stage.oaro_feed_water_flux_con = Constraint(
            expr=(
                0.1 / 5,
                pyunits.convert(
                    stage.flux_mass_phase_comp[0, 0, "Liq", "H2O"] / stage.dens_solvent,
                    to_units=pyunits.L / pyunits.m**2 / pyunits.hr,
                ),
                10 * 1.5,
            )
        )

        stage.oaro_permeate_water_flux_con = Constraint(
            expr=(
                0.1 / 5,
                pyunits.convert(
                    stage.flux_mass_phase_comp[0, 1, "Liq", "H2O"] / stage.dens_solvent,
                    to_units=pyunits.L / pyunits.m**2 / pyunits.hr,
                ),
                10 * 1.5,
            )
        )

        stage.oaro_avg_salt_flux_con = Constraint(
            expr=(
                0,
                pyunits.convert(
                    stage.flux_mass_phase_comp_avg[0, "Liq", "NaCl"],
                    to_units=pyunits.g / pyunits.m**2 / pyunits.hr,
                ),
                50,
            )
        )

        stage.oaro_inlet_salt_flux_con = Constraint(
            expr=(
                0,
                pyunits.convert(
                    stage.flux_mass_phase_comp[0, 0, "Liq", "NaCl"],
                    to_units=pyunits.g / pyunits.m**2 / pyunits.hr,
                ),
                50,
            )
        )

        stage.oaro_permeate_salt_flux_con = Constraint(
            expr=(
                0,
                pyunits.convert(
                    stage.flux_mass_phase_comp[0, 1, "Liq", "NaCl"],
                    to_units=pyunits.g / pyunits.m**2 / pyunits.hr,
                ),
                50,
            )
        )

        stage.min_permeate_flow_rate_con = Constraint(
            expr=sum(
                stage.permeate_inlet.flow_mass_phase_comp[0, "Liq", j]
                for j in ["H2O", "NaCl"]
            )
            >= 0.15
            * sum(
                stage.feed_inlet.flow_mass_phase_comp[0, "Liq", j]
                for j in ["H2O", "NaCl"]
            )
        )

        stage.max_permeate_flow_rate_con = Constraint(
            expr=sum(
                stage.permeate_inlet.flow_mass_phase_comp[0, "Liq", j]
                for j in ["H2O", "NaCl"]
            )
            <= 0.8
            * sum(
                stage.feed_inlet.flow_mass_phase_comp[0, "Liq", j]
                for j in ["H2O", "NaCl"]
            )
        )

    # RO
    m.fs.RO.recovery_mass_phase_comp[0, "Liq", "H2O"].unfix()

    m.fs.RO.area.setlb(1)
    m.fs.RO.area.setub(20000)

    m.fs.RO.width.unfix()
    m.fs.RO.width.setlb(0.1)
    m.fs.RO.width.setub(1000)

    m.fs.RO.feed_side.N_Re[0, 0].setlb(100)
    m.fs.RO.feed_side.N_Re[0, 0].setub(2000)

    m.fs.RO.feed_side.N_Re[0, 1].setlb(100)
    m.fs.RO.feed_side.N_Re[0, 1].setub(2000)

    m.fs.RO.feed_side.properties[0, 0].conc_mass_phase_comp["Liq", "NaCl"].setlb(
        10 * pyunits.g / pyunits.L
    )
    m.fs.RO.feed_side.properties[0, 1].conc_mass_phase_comp["Liq", "NaCl"].setlb(
        10 * pyunits.g / pyunits.L
    )

    m.fs.RO.ro_avg_water_flux_con = Constraint(
        expr=(
            0.4,
            pyunits.convert(
                m.fs.RO.flux_mass_phase_comp_avg[0, "Liq", "H2O"]
                / m.fs.RO.dens_solvent,
                to_units=pyunits.L / pyunits.m**2 / pyunits.hr,
            ),
            40,
        )
    )

    m.fs.RO.ro_feed_water_flux_con = Constraint(
        expr=(
            0.4 / 5,
            pyunits.convert(
                m.fs.RO.flux_mass_phase_comp[0.0, 0.0, "Liq", "H2O"]
                / m.fs.RO.dens_solvent,
                to_units=pyunits.L / pyunits.m**2 / pyunits.hr,
            ),
            40 * 1.5,
        )
    )

    m.fs.RO.ro_permeate_water_flux_con = Constraint(
        expr=(
            0.4 / 5,
            pyunits.convert(
                m.fs.RO.flux_mass_phase_comp[0.0, 1.0, "Liq", "H2O"]
                / m.fs.RO.dens_solvent,
                to_units=pyunits.L / pyunits.m**2 / pyunits.hr,
            ),
            40 * 1.5,
        )
    )

    m.fs.RO.ro_avg_salt_flux_con = Constraint(
        expr=(
            0,
            pyunits.convert(
                m.fs.RO.flux_mass_phase_comp_avg[0, "Liq", "NaCl"],
                to_units=pyunits.g / pyunits.m**2 / pyunits.hr,
            ),
            50,
        )
    )

    m.fs.RO.ro_feed_salt_flux_con = Constraint(
        expr=(
            0,
            pyunits.convert(
                m.fs.RO.flux_mass_phase_comp[0, 0, "Liq", "NaCl"],
                to_units=pyunits.g / pyunits.m**2 / pyunits.hr,
            ),
            50,
        )
    )

    m.fs.RO.ro_retentate_salt_flux_con = Constraint(
        expr=(
            0,
            pyunits.convert(
                m.fs.RO.flux_mass_phase_comp[0, 1, "Liq", "NaCl"],
                to_units=pyunits.g / pyunits.m**2 / pyunits.hr,
            ),
            50,
        )
    )

    # Separator
    for sep in m.fs.Separators.values():
        sep.split_fraction[:, "treat"].unfix()
        sep.split_fraction[:, "treat"].setlb(0)
        sep.split_fraction[:, "treat"].setub(1)

    # Makeup
    if number_of_stages > 1:
        for i in range(2, number_of_stages + 1):
            m.fs.ProductSeparator.split_fraction[:, "makeup" + str(i)].unfix()
            m.fs.ProductSeparator.split_fraction[:, "makeup" + str(i)].setlb(0)
            m.fs.ProductSeparator.split_fraction[:, "makeup" + str(i)].setub(1)

    # additional specifications
    m.fs.product_salinity = Param(
        initialize=500e-6, mutable=True
    )  # product NaCl mass fraction [-]

    # additional constraints
    if water_recovery is not None:
        # product mass flow rate fraction of feed [-]
        m.fs.mass_water_recovery.fix(water_recovery)

    m.fs.eq_product_quality = Constraint(
        expr=m.fs.product.properties[0].mass_frac_phase_comp["Liq", "NaCl"]
        <= m.fs.product_salinity
    )
    iscale.constraint_scaling_transform(
        m.fs.eq_product_quality, 1e3
    )  # scaling constraint

    if number_of_stages > 1:
        m.fs.purge_rate_con = Constraint(
            expr=(
                0,
                sum(
                    m.fs.Separators[stage].purge.flow_mass_phase_comp[0, "Liq", j]
                    for stage in m.fs.NonFirstStages
                    for j in ["H2O", "NaCl"]
                )
                / (
                    sum(
                        m.fs.OAROUnits[1].permeate_outlet.flow_mass_phase_comp[
                            0, "Liq", j
                        ]
                        for j in ["H2O", "NaCl"]
                    )
                    - sum(
                        m.fs.OAROUnits[1].permeate_inlet.flow_mass_phase_comp[
                            0, "Liq", j
                        ]
                        for j in ["H2O", "NaCl"]
                    )
                ),
                0.2,
            )
        )

        assert_degrees_of_freedom(
            m,
            6 * (m.fs.NumberOfStages - 1)
            + 3
            - (1 if (water_recovery is not None) else 0),
        )


def display_system(m):
    print("----system metrics----")
    feed_flow_vol = m.fs.feed.properties[0].flow_vol_phase["Liq"].value * 3600
    feed_flow_mass = sum(
        m.fs.feed.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
    )
    feed_mass_frac_NaCl = (
        m.fs.feed.flow_mass_phase_comp[0, "Liq", "NaCl"].value / feed_flow_mass
    )
    print(
        "Feed: %.2f m3/h, %.2f kg/s, %.0f ppm"
        % (feed_flow_vol, feed_flow_mass, feed_mass_frac_NaCl * 1e6)
    )

    prod_flow_mass = sum(
        m.fs.product.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
    )
    prod_mass_frac_NaCl = (
        m.fs.product.flow_mass_phase_comp[0, "Liq", "NaCl"].value / prod_flow_mass
    )
    print("Product: %.3f kg/s, %.0f ppm" % (prod_flow_mass, prod_mass_frac_NaCl * 1e6))

    brine_flow_mass = sum(
        m.fs.disposal.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
    )
    brine_mass_frac_NaCl = (
        m.fs.disposal.flow_mass_phase_comp[0, "Liq", "NaCl"].value / brine_flow_mass
    )
    print("Brine: %.3f kg/s, %.0f ppm" % (brine_flow_mass, brine_mass_frac_NaCl * 1e6))
    print("Mass water recovery: %.1f%%" % (value(m.fs.mass_water_recovery) * 100))
    print("Volumetric water recovery: %.1f%%" % (value(m.fs.water_recovery) * 100))
    print(f"Number of Stages: {value(m.fs.NumberOfStages)}")
    total_area = value(
        sum(m.fs.OAROUnits[a].area for a in m.fs.NonFinalStages) + m.fs.RO.area
    )
    print(f"Total Membrane Area: {total_area:.2f}")
    print(
        "Energy Consumption: %.1f kWh/m3"
        % value(m.fs.costing.specific_energy_consumption)
    )

    print("Levelized cost of water: %.2f $/m3" % value(m.fs.costing.LCOW))
    print(
        f"Primary Pump Capital Cost ($/m3):"
        f"{value(m.fs.costing.factor_capital_annualization*sum(m.fs.PrimaryPumps[stage].costing.capital_cost for stage in m.fs.Stages)/ m.fs.costing.annual_water_production)}"
    )
    print(
        f"Recycle Pump Capital Cost ($/m3): "
        f"{value(m.fs.costing.factor_capital_annualization*sum(m.fs.RecyclePumps[stage].costing.capital_cost for stage in m.fs.NonFirstStages) / m.fs.costing.annual_water_production)}"
    )
    print(
        f"ERD Capital Cost ($/m3):"
        f"{value(m.fs.costing.factor_capital_annualization*sum(erd.costing.capital_cost for erd in m.fs.EnergyRecoveryDevices.values()) / m.fs.costing.annual_water_production)}"
    )
    print(
        f"Membrane Capital Cost ($/m3): "
        f"{value(m.fs.costing.factor_capital_annualization*(sum(m.fs.OAROUnits[stage].costing.capital_cost for stage in m.fs.NonFinalStages) + m.fs.RO.costing.capital_cost) / m.fs.costing.annual_water_production)}"
    )
    print(
        f"Indirect Capital Cost ($/m3): "
        f"{value(m.fs.costing.factor_capital_annualization*(m.fs.costing.total_capital_cost - m.fs.costing.aggregate_capital_cost) / m.fs.costing.annual_water_production)}"
    )
    electricity_cost = value(
        m.fs.costing.aggregate_flow_costs["electricity"]
        * m.fs.costing.utilization_factor
        / m.fs.costing.annual_water_production
    )
    print(f"Electricity cost ($/m3): {electricity_cost}")

    print("\n")


def display_design(m):
    print("--decision variables--")
    for stage in m.fs.NonFinalStages:
        print(
            "OARO Stage %d water recovery: %.3f"
            % (
                stage,
                value(m.fs.OAROUnits[stage].recovery_mass_phase_comp[0, "Liq", "H2O"]),
            )
        )
        print(
            "OARO Stage %d average water flux: %.1f L/m2/h"
            % (
                stage,
                value(m.fs.OAROUnits[stage].flux_mass_phase_comp_avg[0, "Liq", "H2O"])
                / 1e3
                * 1000
                * 3600,
            )
        )
        print(
            "OARO Stage %d average salt flux: %.1f g/m2/h"
            % (
                stage,
                value(m.fs.OAROUnits[stage].flux_mass_phase_comp_avg[0, "Liq", "NaCl"])
                * 1000
                * 3600,
            )
        )
        print(
            "OARO Stage %d feed operating pressure %.1f bar"
            % (stage, m.fs.OAROUnits[stage].feed_inlet.pressure[0].value / 1e5)
        )
        print(
            "OARO Stage %d permeate operating pressure %.1f bar"
            % (stage, m.fs.OAROUnits[stage].permeate_inlet.pressure[0].value / 1e5)
        )
        print(
            "OARO Stage %d membrane area      %.1f m2"
            % (stage, m.fs.OAROUnits[stage].area.value)
        )
        print(
            "OARO Stage %d water perm. coeff.  %.3f LMH/bar"
            % (stage, m.fs.OAROUnits[stage].A_comp[0, "H2O"].value * (3.6e11))
        )
        print(
            "OARO Stage %d salt perm. coeff.  %.3f LMH/bar"
            % (stage, m.fs.OAROUnits[stage].B_comp[0, "NaCl"].value * (1000.0 * 3600.0))
        )
        print(
            "OARO Stage %d feed-side velocity.  %.3f"
            % (stage, m.fs.OAROUnits[stage].feed_side.velocity[0, 0].value)
        )
        print(
            "OARO Stage %d feed-side Reynolds number.  %.3f"
            % (stage, m.fs.OAROUnits[stage].feed_side.N_Re[0, 0].value)
        )
    print(
        "RO water recovery: %.3f"
        % (value(m.fs.RO.recovery_mass_phase_comp[0, "Liq", "H2O"]))
    )
    print(
        "RO feed operating pressure %.1f bar" % (m.fs.RO.inlet.pressure[0].value / 1e5)
    )
    print(
        "RO permeate operating pressure %.1f bar"
        % (m.fs.RO.permeate.pressure[0].value / 1e5)
    )
    print("RO membrane area      %.1f m2" % (m.fs.RO.area.value))
    print("RO membrane width      %.1f m" % (m.fs.RO.width.value))
    print(
        "RO water perm. coeff.  %.3f LMH/bar"
        % (m.fs.RO.A_comp[0, "H2O"].value * (3.6e11))
    )
    print(
        "RO salt perm. coeff.  %.3f LMH"
        % (m.fs.RO.B_comp[0, "NaCl"].value * (1000.0 * 3600.0))
    )


def display_state(m):
    print("--------state---------")

    def print_state(s, b):
        feed_flow_mass = sum(
            m.fs.feed.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
        )
        flow_mass = sum(
            b.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
        )
        normalized_flow_mass = flow_mass / feed_flow_mass * 100
        mass_frac_ppm = b.flow_mass_phase_comp[0, "Liq", "NaCl"].value / flow_mass * 1e3
        pressure_bar = b.pressure[0].value / 1e5
        print(
            s.ljust(20)
            + ": %.0f, %.3f g/L, %.1f bar, %.4f kg/s, %.4f kg/s"
            % (
                normalized_flow_mass,
                mass_frac_ppm,
                pressure_bar,
                b.flow_mass_phase_comp[0, "Liq", "H2O"].value,
                b.flow_mass_phase_comp[0, "Liq", "NaCl"].value,
            )
        )

    print_state("Feed", m.fs.feed.outlet)

    for stage in m.fs.Stages:

        print_state(f"Primary Pump {stage} out", m.fs.PrimaryPumps[stage].outlet)
        print_state(
            f"ERD {stage} out",
            m.fs.EnergyRecoveryDevices[stage].outlet,
        )

        if stage == m.fs.LastStage:
            pass
        else:
            print_state(f"OARO {stage} feed inlet", m.fs.OAROUnits[stage].feed_inlet)
            print_state(f"OARO {stage} feed outlet", m.fs.OAROUnits[stage].feed_outlet)
            print_state(
                f"OARO {stage} permeate inlet", m.fs.OAROUnits[stage].permeate_inlet
            )
            print_state(
                f"OARO {stage} permeate outlet", m.fs.OAROUnits[stage].permeate_outlet
            )

        if stage == m.fs.FirstStage:
            pass
        else:
            print_state(f"Recycle Pump {stage} out", m.fs.RecyclePumps[stage].outlet)
            print_state(f"Purge {stage}", m.fs.Separators[stage].purge)
            print_state(f"Make-up {stage}", m.fs.Mixers[stage].makeup)

    print_state(f"RO inlet", m.fs.RO.inlet)
    print_state(f"RO permeate", m.fs.RO.permeate)
    print_state(f"RO retentate", m.fs.RO.retentate)

    print_state(f"Disposal", m.fs.disposal.inlet)
    print_state(f"Product", m.fs.product.inlet)


if __name__ == "__main__":
    m = main(number_of_stages=3, system_recovery=0.5, erd_type=ERDtype.pump_as_turbine)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Initial property package for pure water system (vapor or liquid)
"""

# Import Python libraries
import idaes.logger as idaeslog

# Import Pyomo libraries
from pyomo.environ import (
    Constraint,
    Expression,
    Reals,
    NonNegativeReals,
    Var,
    Param,
    Suffix,
    value,
    log,
    exp,
    check_optimal_termination,
)
from pyomo.environ import units as pyunits

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    PhysicalParameterBlock,
    StateBlockData,
    StateBlock,
    MaterialBalanceType,
    EnergyBalanceType,
)
from idaes.core.base.components import Component
from idaes.core.base.phases import LiquidPhase, VaporPhase
from idaes.core.util.initialization import (
    fix_state_vars,
    revert_state_vars,
    solve_indexed_blocks,
)
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_unfixed_variables,
)
from idaes.core.util.exceptions import (
    ConfigurationError,
    InitializationError,
    PropertyPackageError,
)
import idaes.core.util.scaling as iscale
from watertap.core.util.scaling import transform_property_constraints

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("WaterParameterBlock")
class WaterParameterData(PhysicalParameterBlock):
    """Parameter block for a water property package."""

    CONFIG = PhysicalParameterBlock.CONFIG()

    def build(self):
        """
        Callable method for Block construction.
        """
        super(WaterParameterData, self).build()

        self._state_block_class = WaterStateBlock

        # components
        self.H2O = Component()

        # phases
        self.Liq = LiquidPhase()
        self.Vap = VaporPhase()

        """ References
        This package was developed from the following references:
        - K.G.Nayar, M.H.Sharqawy, L.D.Banchik, and J.H.Lienhard V, "Thermophysical properties of seawater: A review and
        new correlations that include pressure dependence,"Desalination, Vol.390, pp.1 - 24, 2016.
        doi: 10.1016/j.desal.2016.02.024(preprint)
        - Mostafa H.Sharqawy, John H.Lienhard V, and Syed M.Zubair, "Thermophysical properties of seawater: A review of
        existing correlations and data,"Desalination and Water Treatment, Vol.16, pp.354 - 380, April 2010.
        (2017 corrections provided at http://web.mit.edu/seawater)
        """

        # Parameters
        # molecular weight
        self.mw_comp = Param(
            self.component_list,
            initialize=18.01528e-3,
            units=pyunits.kg / pyunits.mol,
            doc="Molecular weight",
        )

        # Liq mass density parameters, eq. 8 in Sharqawy et al. (2010)
        dens_units = pyunits.kg / pyunits.m**3
        t_inv_units = pyunits.K**-1
        s_inv_units = pyunits.kg / pyunits.g

        self.dens_mass_param_A1 = Var(
            within=Reals,
            initialize=9.999e2,
            units=dens_units,
            doc="Mass density parameter A1",
        )
        self.dens_mass_param_A2 = Var(
            within=Reals,
            initialize=2.034e-2,
            units=dens_units * t_inv_units,
            doc="Mass density parameter A2",
        )
        self.dens_mass_param_A3 = Var(
            within=Reals,
            initialize=-6.162e-3,
            units=dens_units * t_inv_units**2,
            doc="Mass density parameter A3",
        )
        self.dens_mass_param_A4 = Var(
            within=Reals,
            initialize=2.261e-5,
            units=dens_units * t_inv_units**3,
            doc="Mass density parameter A4",
        )
        self.dens_mass_param_A5 = Var(
            within=Reals,
            initialize=-4.657e-8,
            units=dens_units * t_inv_units**4,
            doc="Mass density parameter A5",
        )

        # Vap mass density parameters (approximating using ideal gas)
        self.dens_mass_param_mw = Var(
            within=Reals,
            initialize=18.01528e-3,
            units=pyunits.kg / pyunits.mol,
            doc="Mass density parameter molecular weight",
        )
        self.dens_mass_param_R = Var(
            within=Reals,
            initialize=8.31462618,
            units=pyunits.J / pyunits.mol / pyunits.K,
            doc="Mass density parameter universal gas constant",
        )

        # vapor pressure parameters,  eq. 5 and 6 in Nayar et al.(2016)
        self.pressure_sat_param_psatw_A1 = Var(
            within=Reals,
            initialize=-5.8002206e3,
            units=pyunits.K,
            doc="Vapor pressure of pure water parameter A1",
        )
        self.pressure_sat_param_psatw_A2 = Var(
            within=Reals,
            initialize=1.3914993,
            units=pyunits.dimensionless,
            doc="Vapor pressure of pure water parameter A2",
        )
        self.pressure_sat_param_psatw_A3 = Var(
            within=Reals,
            initialize=-4.8640239e-2,
            units=t_inv_units,
            doc="Vapor pressure of pure water parameter A3",
        )
        self.pressure_sat_param_psatw_A4 = Var(
            within=Reals,
            initialize=4.1764768e-5,
            units=t_inv_units**2,
            doc="Vapor pressure of pure water parameter A4",
        )
        self.pressure_sat_param_psatw_A5 = Var(
            within=Reals,
            initialize=-1.4452093e-8,
            units=t_inv_units**3,
            doc="Vapor pressure of pure water parameter A5",
        )
        self.pressure_sat_param_psatw_A6 = Var(
            within=Reals,
            initialize=6.5459673,
            units=pyunits.dimensionless,
            doc="Vapor pressure of pure water parameter A6",
        )

        # specific enthalpy parameters, Table 9 in Nayar et al. (2016) - ignores salinity-based parameters
        enth_mass_units = pyunits.J / pyunits.kg
        P_inv_units = pyunits.MPa**-1

        self.enth_mass_param_A1 = Var(
            within=Reals,
            initialize=996.7767,
            units=enth_mass_units * P_inv_units,
            doc="Specific enthalpy parameter A1",
        )
        self.enth_mass_param_A2 = Var(
            within=Reals,
            initialize=-3.2406,
            units=enth_mass_units * P_inv_units * t_inv_units,
            doc="Specific enthalpy parameter A2",
        )
        self.enth_mass_param_A3 = Var(
            within=Reals,
            initialize=0.0127,
            units=enth_mass_units * P_inv_units * t_inv_units**2,
            doc="Specific enthalpy parameter A3",
        )
        self.enth_mass_param_A4 = Var(
            within=Reals,
            initialize=-4.7723e-5,
            units=enth_mass_units * P_inv_units * t_inv_units**3,
            doc="Specific enthalpy parameter A4",
        )
        self.enth_mass_param_C1 = Var(
            within=Reals,
            initialize=141.355,
            units=enth_mass_units,
            doc="Specific enthalpy parameter C1",
        )
        self.enth_mass_param_C2 = Var(
            within=Reals,
            initialize=4202.07,
            units=enth_mass_units * t_inv_units,
            doc="Specific enthalpy parameter C2",
        )
        self.enth_mass_param_C3 = Var(
            within=Reals,
            initialize=-0.535,
            units=enth_mass_units * t_inv_units**2,
            doc="Specific enthalpy parameter C3",
        )
        self.enth_mass_param_C4 = Var(
            within=Reals,
            initialize=0.004,
            units=enth_mass_units * t_inv_units**3,
            doc="Specific enthalpy parameter C4",
        )
        # specific heat parameters from eq (9) in Sharqawy et al. (2010)
        cp_units = pyunits.J / (pyunits.kg * pyunits.K)
        self.cp_phase_param_A1 = Var(
            within=Reals,
            initialize=5.328,
            units=cp_units,
            doc="Specific heat of seawater parameter A1",
        )
        self.cp_phase_param_B1 = Var(
            within=Reals,
            initialize=-6.913e-3,
            units=cp_units * t_inv_units,
            doc="Specific heat of seawater parameter B1",
        )
        self.cp_phase_param_C1 = Var(
            within=Reals,
            initialize=9.6e-6,
            units=cp_units * t_inv_units**2,
            doc="Specific heat of seawater parameter C1",
        )
        self.cp_phase_param_D1 = Var(
            within=Reals,
            initialize=2.5e-9,
            units=cp_units * t_inv_units**3,
            doc="Specific heat of seawater parameter D1",
        )

        # Specific heat parameters for Cp vapor from NIST Webbook
        # Chase, M.W., Jr., NIST-JANAF Themochemical Tables, Fourth Edition, J. Phys. Chem. Ref. Data, Monograph 9, 1998, 1-1951
        self.cp_vap_param_A = Var(
            within=Reals,
            initialize=30.09200 / 18.01528e-3,
            units=cp_units,
            doc="Specific heat of water vapor parameter A",
        )
        self.cp_vap_param_B = Var(
            within=Reals,
            initialize=6.832514 / 18.01528e-3,
            units=cp_units * t_inv_units,
            doc="Specific heat of water vapor parameter B",
        )
        self.cp_vap_param_C = Var(
            within=Reals,
            initialize=6.793435 / 18.01528e-3,
            units=cp_units * t_inv_units**2,
            doc="Specific heat of water vapor parameter C",
        )
        self.cp_vap_param_D = Var(
            within=Reals,
            initialize=-2.534480 / 18.01528e-3,
            units=cp_units * t_inv_units**3,
            doc="Specific heat of water vapor parameter D",
        )
        self.cp_vap_param_E = Var(
            within=Reals,
            initialize=0.082139 / 18.01528e-3,
            units=cp_units * t_inv_units**-2,
            doc="Specific heat of water vapor parameter E",
        )

        # latent heat of pure water parameters from eq. 54 in Sharqawy et al. (2010)
        self.dh_vap_w_param_0 = Var(
            within=Reals,
            initialize=2.501e6,
            units=enth_mass_units,
            doc="Latent heat of pure water parameter 0",
        )
        self.dh_vap_w_param_1 = Var(
            within=Reals,
            initialize=-2.369e3,
            units=cp_units,
            doc="Latent heat of pure water parameter 1",
        )
        self.dh_vap_w_param_2 = Var(
            within=Reals,
            initialize=2.678e-1,
            units=enth_mass_units * t_inv_units**2,
            doc="Latent heat of pure water parameter 2",
        )
        self.dh_vap_w_param_3 = Var(
            within=Reals,
            initialize=-8.103e-3,
            units=enth_mass_units * t_inv_units**3,
            doc="Latent heat of pure water parameter 3",
        )
        self.dh_vap_w_param_4 = Var(
            within=Reals,
            initialize=-2.079e-5,
            units=enth_mass_units * t_inv_units**4,
            doc="Latent heat of pure water parameter 4",
        )

        # traditional parameters are the only Vars currently on the block and should be fixed
        for v in self.component_objects(Var):
            v.fix()

        # ---default scaling---
        self.set_default_scaling("temperature", 1e-2)
        self.set_default_scaling("pressure", 1e-5)
        self.set_default_scaling("dens_mass_phase", 1e-3, index="Liq")
        self.set_default_scaling("dens_mass_phase", 1, index="Vap")
        self.set_default_scaling("enth_mass_phase", 1e-5, index="Liq")
        self.set_default_scaling("enth_mass_phase", 1e-6, index="Vap")
        self.set_default_scaling("pressure_sat", 1e-5)
        self.set_default_scaling("cp_mass_phase", 1e-3, index="Liq")
        self.set_default_scaling("cp_mass_phase", 1e-3, index="Vap")
        self.set_default_scaling("dh_vap_mass", 1e-6)

    @classmethod
    def define_metadata(cls, obj):
        """Define properties supported and units."""
        obj.add_properties(
            {
                "flow_mass_phase_comp": {"method": None},
                "temperature": {"method": None},
                "pressure": {"method": None},
                "dens_mass_phase": {"method": "_dens_mass_phase"},
                "flow_vol_phase": {"method": "_flow_vol_phase"},
                "flow_vol": {"method": "_flow_vol"},
                "flow_mol_phase_comp": {"method": "_flow_mol_phase_comp"},
                "mole_frac_phase_comp": {"method": "_mole_frac_phase_comp"},
                "pressure_sat": {"method": "_pressure_sat"},
                "enth_mass_phase": {"method": "_enth_mass_phase"},
                "enth_flow_phase": {"method": "_enth_flow_phase"},
                "cp_mass_phase": {"method": "_cp_mass_phase"},
            }
        )

        obj.define_custom_properties(
            {
                "dh_vap_mass": {"method": "_dh_vap_mass"},
            }
        )

        obj.add_default_units(
            {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


class _WaterStateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    def fix_initialization_states(self):
        """
        Fixes state variables for state blocks.

        Returns:
            None
        """
        # Fix state variables
        fix_state_vars(self)

        # Constraint on water concentration at outlet - unfix in these cases
        for b in self.values():
            if b.config.defined_state is False:
                b.conc_mol_comp["H2O"].unfix()

    def initialize(
        self,
        state_args=None,
        state_vars_fixed=False,
        hold_state=False,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for property package.
        Keyword Arguments:
            state_args : Dictionary with initial guesses for the state vars
                         chosen. Note that if this method is triggered
                         through the control volume, and if initial guesses
                         were not provided at the unit model level, the
                         control volume passes the inlet values as initial
                         guess.The keys for the state_args dictionary are:
                         flow_mass_phase_comp : value at which to initialize
                                               phase component flows
                         pressure : value at which to initialize pressure
                         temperature : value at which to initialize temperature
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default={})
            state_vars_fixed: Flag to denote if state vars have already been
                              fixed.
                              - True - states have already been fixed by the
                                       control volume 1D. Control volume 0D
                                       does not fix the state vars, so will
                                       be False if this state block is used
                                       with 0D blocks.
                             - False - states have not been fixed. The state
                                       block will deal with fixing/unfixing.
            solver : Solver object to use during initialization if None is provided
                     it will use the default solver for IDAES (default = None)
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states variables are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 release_state method
        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """
        # Get loggers
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="properties")

        # Set solver and options
        opt = get_solver(solver, optarg)

        # Fix state variables
        flags = fix_state_vars(self, state_args)
        # Check when the state vars are fixed already result in dof 0
        for k in self.keys():
            dof = degrees_of_freedom(self[k])
            if dof != 0:
                raise PropertyPackageError(
                    "State vars fixed but degrees of "
                    "freedom for state block is not "
                    "zero during initialization."
                )

        # ---------------------------------------------------------------------
        skip_solve = True  # skip solve if only state variables are present
        for k in self.keys():
            if number_unfixed_variables(self[k]) != 0:
                skip_solve = False

        if not skip_solve:
            # Initialize properties
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = solve_indexed_blocks(opt, [self], tee=slc.tee)
            init_log.info_high(
                "Property initialization: {}.".format(idaeslog.condition(results))
            )

        # If input block, return flags, else release state
        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                self.release_state(flags)

        if (not skip_solve) and (not check_optimal_termination(results)):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please "
                f"check the output logs for more information."
            )

    def release_state(self, flags, outlvl=idaeslog.NOTSET):
        """
        Method to release state variables fixed during initialisation.
        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        """
        # Unfix state variables
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        revert_state_vars(self, flags)
        init_log.info("{} State Released.".format(self.name))

    def calculate_state(
        self,
        var_args=None,
        hold_state=False,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Solves state blocks given a set of variables and their values. These variables can
        be state variables or properties. This method is typically used before
        initialization to solve for state variables because non-state variables (i.e. properties)
        cannot be fixed in initialization routines.
        Keyword Arguments:
            var_args : dictionary with variables and their values, they can be state variables or properties
                       {(VAR_NAME, INDEX): VALUE}
            hold_state : flag indicating whether all of the state variables should be fixed after calculate state.
                         True - State variables will be fixed.
                         False - State variables will remain unfixed, unless already fixed.
            outlvl : idaes logger object that sets output level of solve call (default=idaeslog.NOTSET)
            solver : solver name string if None is provided the default solver
                     for IDAES will be used (default = None)
            optarg : solver options dictionary object (default={})
        Returns:
            results object from state block solve
        """
        # Get logger
        solve_log = idaeslog.getSolveLogger(self.name, level=outlvl, tag="properties")

        # Initialize at current state values (not user provided)
        self.initialize(solver=solver, optarg=optarg, outlvl=outlvl)

        # Set solver and options
        opt = get_solver(solver, optarg)

        # Fix variables and check degrees of freedom
        flags = (
            {}
        )  # dictionary noting which variables were fixed and their previous state
        for k in self.keys():
            sb = self[k]
            for (v_name, ind), val in var_args.items():
                var = getattr(sb, v_name)
                if iscale.get_scaling_factor(var[ind]) is None:
                    _log.warning(
                        "While using the calculate_state method on {sb_name}, variable {v_name} "
                        "was provided as an argument in var_args, but it does not have a scaling "
                        "factor. This suggests that the calculate_scaling_factor method has not been "
                        "used or the variable was created on demand after the scaling factors were "
                        "calculated. It is recommended to touch all relevant variables (i.e. call "
                        "them or set an initial value) before using the calculate_scaling_factor "
                        "method.".format(v_name=v_name, sb_name=sb.name)
                    )
                if var[ind].is_fixed():
                    flags[(k, v_name, ind)] = True
                    if value(var[ind]) != val:
                        raise ConfigurationError(
                            "While using the calculate_state method on {sb_name}, {v_name} was "
                            "fixed to a value {val}, but it was already fixed to value {val_2}. "
                            "Unfix the variable before calling the calculate_state "
                            "method or update var_args."
                            "".format(
                                sb_name=sb.name,
                                v_name=var.name,
                                val=val,
                                val_2=value(var[ind]),
                            )
                        )
                else:
                    flags[(k, v_name, ind)] = False
                    var[ind].fix(val)

            if degrees_of_freedom(sb) != 0:
                raise RuntimeError(
                    "While using the calculate_state method on {sb_name}, the degrees "
                    "of freedom were {dof}, but 0 is required. Check var_args and ensure "
                    "the correct fixed variables are provided."
                    "".format(sb_name=sb.name, dof=degrees_of_freedom(sb))
                )

        # Solve
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = solve_indexed_blocks(opt, [self], tee=slc.tee)
            solve_log.info_high(
                "Calculate state: {}.".format(idaeslog.condition(results))
            )

        if not check_optimal_termination(results):
            _log.warning(
                "While using the calculate_state method on {sb_name}, the solver failed "
                "to converge to an optimal solution. This suggests that the user provided "
                "infeasible inputs, or that the model is poorly scaled, poorly initialized, "
                "or degenerate."
            )

        # unfix all variables fixed with var_args
        for (k, v_name, ind), previously_fixed in flags.items():
            if not previously_fixed:
                var = getattr(self[k], v_name)
                var[ind].unfix()

        # fix state variables if hold_state
        if hold_state:
            fix_state_vars(self)

        return results


@declare_process_block_class("WaterStateBlock", block_class=_WaterStateBlock)
class WaterStateBlockData(StateBlockData):
    """A water property package."""

    def build(self):
        """Callable method for Block construction."""
        super().build()

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # Add state variables
        self.flow_mass_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=0.5,
            bounds=(0.0, None),
            domain=NonNegativeReals,
            units=pyunits.kg / pyunits.s,
            doc="Mass flow rate",
        )

        self.temperature = Var(
            initialize=298.15,
            bounds=(273.15, 1000),
            domain=NonNegativeReals,
            units=pyunits.K,
            doc="Temperature",
        )

        self.pressure = Var(
            initialize=101325,
            bounds=(1e3, 5e7),
            domain=NonNegativeReals,
            units=pyunits.Pa,
            doc="Pressure",
        )

    # -----------------------------------------------------------------------------
    # Property Methods

    def _dens_mass_phase(self):
        self.dens_mass_phase = Var(
            self.params.phase_list,
            initialize={"Liq": 1e3, "Vap": 1},
            bounds=(1e-3, 1e6),
            units=pyunits.kg * pyunits.m**-3,
            doc="Mass density of seawater",
        )

        def rule_dens_mass_phase(b, phase):  # density, eq. 8 in Sharqawy
            t = b.temperature - 273.15 * pyunits.K
            # s = b.mass_frac_phase_comp['Liq', 'TDS']
            if phase == "Liq":
                dens_mass = (
                    b.params.dens_mass_param_A1
                    + b.params.dens_mass_param_A2 * t
                    + b.params.dens_mass_param_A3 * t**2
                    + b.params.dens_mass_param_A4 * t**3
                    + b.params.dens_mass_param_A5 * t**4
                )
                return b.dens_mass_phase[phase] == dens_mass
            else:  # phase == 'Vap'
                dens_mass = (
                    b.params.dens_mass_param_mw
                    * b.pressure
                    / b.params.dens_mass_param_R
                    / b.temperature
                )
                return b.dens_mass_phase[phase] == dens_mass

        self.eq_dens_mass_phase = Constraint(
            self.params.phase_list, rule=rule_dens_mass_phase
        )

    def _flow_vol_phase(self):
        self.flow_vol_phase = Var(
            self.params.phase_list,
            initialize=1,
            bounds=(None, None),
            units=pyunits.m**3 / pyunits.s,
            doc="Volumetric flow rate",
        )

        def rule_flow_vol_phase(b, p):
            return (
                b.flow_vol_phase[p]
                == b.flow_mass_phase_comp[p, "H2O"] / b.dens_mass_phase[p]
            )

        self.eq_flow_vol_phase = Constraint(
            self.params.phase_list, rule=rule_flow_vol_phase
        )

    def _flow_vol(self):
        def rule_flow_vol(b):
            return sum(b.flow_vol_phase[p] for p in b.params.phase_list)

        self.flow_vol = Expression(rule=rule_flow_vol)

    def _flow_mol_phase_comp(self):
        self.flow_mol_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=100,
            bounds=(None, None),
            units=pyunits.mol / pyunits.s,
            doc="Molar flowrate",
        )

        def rule_flow_mol_phase_comp(b, p, j):
            return (
                b.flow_mol_phase_comp[p, j]
                == b.flow_mass_phase_comp[p, j] / b.params.mw_comp[j]
            )

        self.eq_flow_mol_phase_comp = Constraint(
            self.params.phase_list,
            self.params.component_list,
            rule=rule_flow_mol_phase_comp,
        )

    def _mole_frac_phase_comp(self):
        self.mole_frac_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=0.1,
            bounds=(0.0, None),
            units=pyunits.dimensionless,
            doc="Mole fraction",
        )

        def rule_mole_frac_phase_comp(b, p, j):
            return b.mole_frac_phase_comp[p, j] == b.flow_mol_phase_comp[p, j] / sum(
                b.flow_mol_phase_comp[p, j] for p in b.params.phase_list
            )

        self.eq_mole_frac_phase_comp = Constraint(
            self.params.phase_list,
            self.params.component_list,
            rule=rule_mole_frac_phase_comp,
        )

    def _enth_mass_phase(self):
        self.enth_mass_phase = Var(
            self.params.phase_list,  # ['Liq','Vap']
            initialize=1e6,
            bounds=(1, 1e9),
            units=pyunits.J * pyunits.kg**-1,
            doc="Specific enthalpy",
        )

        def rule_enth_mass_phase(
            b, p
        ):  # specific enthalpy, eq. 25 and 26 in Nayar et al. (2016)
            t = (
                b.temperature - 273.15 * pyunits.K
            )  # temperature in degC, but pyunits in K
            P = b.pressure - 101325 * pyunits.Pa
            P_MPa = pyunits.convert(P, to_units=pyunits.MPa)

            # water enthalpy without pressure effects
            h_w = (
                b.params.enth_mass_param_C1
                + b.params.enth_mass_param_C2 * t
                + b.params.enth_mass_param_C3 * t**2
                + b.params.enth_mass_param_C4 * t**3
            )

            # water enthalpy with pressure effects
            h_w_P = h_w + P_MPa * (
                b.params.enth_mass_param_A1
                + b.params.enth_mass_param_A2 * t
                + b.params.enth_mass_param_A3 * t**2
                + b.params.enth_mass_param_A4 * t**3
            )

            if p == "Liq":
                return b.enth_mass_phase[p] == h_w_P
            else:
                return b.enth_mass_phase[p] == h_w_P + b.dh_vap_mass

        self.eq_enth_mass_phase = Constraint(
            self.params.phase_list, rule=rule_enth_mass_phase
        )

    def _pressure_sat(self):
        self.pressure_sat = Var(
            initialize=1e3,
            bounds=(1, 1e8),
            units=pyunits.Pa,
            doc="Saturation vapor pressure",
        )

        # vapor pressure, eq. 5 and 6 in Nayar et al.(2016)
        def rule_pressure_sat(b):
            t = b.temperature
            psatw = (
                exp(
                    b.params.pressure_sat_param_psatw_A1 * t**-1
                    + b.params.pressure_sat_param_psatw_A2
                    + b.params.pressure_sat_param_psatw_A3 * t
                    + b.params.pressure_sat_param_psatw_A4 * t**2
                    + b.params.pressure_sat_param_psatw_A5 * t**3
                    + b.params.pressure_sat_param_psatw_A6 * log(t / pyunits.K)
                )
                * pyunits.Pa
            )
            return b.pressure_sat == psatw

        self.eq_pressure_sat = Constraint(rule=rule_pressure_sat)

    def _enth_flow_phase(self):
        # enthalpy flow variable
        self.enth_flow_phase = Var(
            self.params.phase_list,  # ['Liq','Vap']
            initialize=1e6,
            bounds=(None, None),
            units=pyunits.J / pyunits.s,
            doc="Enthalpy flow",
        )

        # enthalpy flow [J/s]
        def rule_enth_flow_phase(b, p):
            return (
                b.enth_flow_phase[p]
                == b.flow_mass_phase_comp[p, "H2O"] * b.enth_mass_phase[p]
            )

        self.eq_enth_flow_phase = Constraint(
            self.params.phase_list, rule=rule_enth_flow_phase
        )

    def _cp_mass_phase(self):
        self.cp_mass_phase = Var(
            self.params.phase_list,
            initialize=4e3,
            bounds=(0.0, 1e8),
            units=pyunits.J / pyunits.kg / pyunits.K,
            doc="Specific heat capacity",
        )

        def rule_cp_mass_phase(b, phase):
            if phase == "Liq":
                # specific heat, eq. 9 in Sharqawy et al. (2010)
                # Convert T90 to T68, eq. 4 in Sharqawy et al. (2010); primary reference from Rusby (1991)
                t = (b.temperature - 0.00025 * 273.15 * pyunits.K) / (1 - 0.00025)
                # s = b.mass_frac_phase_comp['Liq', 'TDS'] * 1000 * pyunits.g / pyunits.kg
                A = (
                    b.params.cp_phase_param_A1
                )  # + b.params.cp_phase_param_A2 * s + b.params.cp_phase_param_A3 * s ** 2
                B = (
                    b.params.cp_phase_param_B1
                )  # + b.params.cp_phase_param_B2 * s + b.params.cp_phase_param_B3 * s ** 2
                C = (
                    b.params.cp_phase_param_C1
                )  # + b.params.cp_phase_param_C2 * s + b.params.cp_phase_param_C3 * s ** 2
                D = (
                    b.params.cp_phase_param_D1
                )  # + b.params.cp_phase_param_D2 * s + b.params.cp_phase_param_D3 * s ** 2
                return (
                    b.cp_mass_phase[phase]
                    == (A + B * t + C * t**2 + D * t**3) * 1000
                )
            # phase == 'Vap'
            else:
                t = b.temperature / 1000
                return (
                    b.cp_mass_phase[phase]
                    == b.params.cp_vap_param_A
                    + b.params.cp_vap_param_B * t
                    + b.params.cp_vap_param_C * t**2
                    + b.params.cp_vap_param_D * t**3
                    + b.params.cp_vap_param_E / t**2
                )

        self.eq_cp_mass_phase = Constraint(
            self.params.phase_list, rule=rule_cp_mass_phase
        )

    def _dh_vap_mass(self):
        self.dh_vap_mass = Var(
            initialize=2.4e6,
            bounds=(1, 1e9),
            units=pyunits.J / pyunits.kg,
            doc="Latent heat of vaporization",
        )

        # latent heat of seawater from eq. 37 and eq. 55 in Sharqawy et al. (2010)
        def rule_dh_vap_mass(b):
            t = b.temperature - 273.15 * pyunits.K
            return (
                b.dh_vap_mass
                == b.params.dh_vap_w_param_0
                + b.params.dh_vap_w_param_1 * t
                + b.params.dh_vap_w_param_2 * t**2
                + b.params.dh_vap_w_param_3 * t**3
                + b.params.dh_vap_w_param_4 * t**4
            )

        self.eq_dh_vap_mass = Constraint(rule=rule_dh_vap_mass)

    # General Methods
    # NOTE: For scaling in the control volume to work properly, these methods must
    # return a pyomo Var or Expression

    def get_material_flow_terms(self, p, j):
        """Create material flow terms for control volume."""
        return self.flow_mass_phase_comp[p, j]

    def get_enthalpy_flow_terms(self, p):
        """Create enthalpy flow terms."""
        return self.enth_flow_phase[p]

    # TODO: make property package compatible with dynamics
    # def get_material_density_terms(self, p, j):
    #     """Create material density terms."""

    # def get_enthalpy_density_terms(self, p):
    #     """Create enthalpy density terms."""

    def default_material_balance_type(self):
        return MaterialBalanceType.componentTotal

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def get_material_flow_basis(self):
        return MaterialFlowBasis.mass

    def define_state_vars(self):
        """Define state vars."""
        return {
            "flow_mass_phase_comp": self.flow_mass_phase_comp,
            "temperature": self.temperature,
            "pressure": self.pressure,
        }

    # -----------------------------------------------------------------------------
    # Scaling methods
    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # setting scaling factors for variables

        # default scaling factors have already been set with
        # idaes.core.property_base.calculate_scaling_factors()
        # for the following variables: flow_mass_phase_comp, pressure,
        # temperature, dens_mass_phase, visc_d_phase, osm_coeff, and enth_mass_phase

        # these variables should have user input
        if iscale.get_scaling_factor(self.flow_mass_phase_comp["Liq", "H2O"]) is None:
            sf = iscale.get_scaling_factor(
                self.flow_mass_phase_comp["Liq", "H2O"], default=1e0, warning=True
            )
            iscale.set_scaling_factor(self.flow_mass_phase_comp["Liq", "H2O"], sf)

        if iscale.get_scaling_factor(self.flow_mass_phase_comp["Vap", "H2O"]) is None:
            sf = iscale.get_scaling_factor(
                self.flow_mass_phase_comp["Vap", "H2O"], default=1e0, warning=True
            )
            iscale.set_scaling_factor(self.flow_mass_phase_comp["Vap", "H2O"], sf)

        # scaling factors for parameters
        for j, v in self.params.mw_comp.items():
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(self.params.mw_comp, 1e2)

        # these variables do not typically require user input,
        # will not override if the user does provide the scaling factor
        if self.is_property_constructed("flow_vol_phase"):
            for p in self.params.phase_list:
                if iscale.get_scaling_factor(self.flow_vol_phase[p]) is None:
                    sf = iscale.get_scaling_factor(
                        self.flow_mass_phase_comp[p, "H2O"]
                    ) / iscale.get_scaling_factor(self.dens_mass_phase[p])
                    iscale.set_scaling_factor(self.flow_vol_phase[p], sf)

        if self.is_property_constructed("flow_vol"):
            if iscale.get_scaling_factor(self.flow_vol) is None:
                sf_liq = iscale.get_scaling_factor(self.flow_vol_phase["Liq"])
                sf_vap = iscale.get_scaling_factor(self.flow_vol_phase["Vap"])
                sf = min(sf_liq, sf_vap)
                iscale.set_scaling_factor(self.flow_vol, sf)

        if self.is_property_constructed("flow_mol_phase_comp"):
            for p in self.params.phase_list:
                if (
                    iscale.get_scaling_factor(self.flow_mol_phase_comp[p, "H2O"])
                    is None
                ):
                    sf = iscale.get_scaling_factor(self.flow_mass_phase_comp[p, "H2O"])
                    sf /= iscale.get_scaling_factor(self.params.mw_comp["H2O"])
                    iscale.set_scaling_factor(self.flow_mol_phase_comp[p, "H2O"], sf)

        if self.is_property_constructed("mole_frac_phase_comp"):
            sf_flow_mol_liq = iscale.get_scaling_factor(
                self.flow_mol_phase_comp["Liq", "H2O"]
            )
            sf_flow_mol_vap = iscale.get_scaling_factor(
                self.flow_mol_phase_comp["Vap", "H2O"]
            )
            sf_flow_mol = min(sf_flow_mol_liq, sf_flow_mol_vap)
            for p in self.params.phase_list:
                if (
                    iscale.get_scaling_factor(self.mole_frac_phase_comp[p, "H2O"])
                    is None
                ):
                    sf = (
                        iscale.get_scaling_factor(self.flow_mol_phase_comp[p, "H2O"])
                        / sf_flow_mol
                    )
                    print("iter:", p, sf)
                    iscale.set_scaling_factor(self.mole_frac_phase_comp[p, "H2O"], sf)

        if self.is_property_constructed("enth_flow_phase"):
            for p in self.params.phase_list:
                if iscale.get_scaling_factor(self.enth_flow_phase[p]) is None:
                    sf = iscale.get_scaling_factor(self.flow_mass_phase_comp[p, "H2O"])
                    sf *= iscale.get_scaling_factor(self.enth_mass_phase[p])
                    iscale.set_scaling_factor(self.enth_flow_phase[p], sf)

        # transforming constraints
        transform_property_constraints(self)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from watertap.ui.fsapi import FlowsheetInterface
from watertap.core.util.initialization import assert_degrees_of_freedom
from watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.dye_desalination.dye_desalination_withRO import (
    build,
    set_operating_conditions,
    initialize_system,
    solve,
    add_costing,
)
from pyomo.environ import units as pyunits, assert_optimal_termination
from pyomo.util.check_units import assert_units_consistent


def export_to_ui():
    return FlowsheetInterface(
        name="Dye Desalination",
        do_export=export_variables,
        do_build=build_flowsheet,
        do_solve=solve_flowsheet,
    )


def export_variables(flowsheet=None, exports=None, build_options=None, **kwargs):
    fs = flowsheet
    # --- Input data ---
    # Feed conditions
    exports.add(
        obj=fs.feed.flow_vol[0],
        name="Volumetric flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=0,
        description="Inlet volumetric flow rate",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "dye"],
        name="Dye concentration",
        ui_units=pyunits.kg / pyunits.m**3,
        display_units="kg/m3",
        rounding=2,
        description="Inlet dye concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "tds"],
        name="TDS concentration",
        ui_units=pyunits.kg / pyunits.m**3,
        display_units="kg/m3",
        rounding=2,
        description="Inlet total dissolved solids (TDS) concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.tb_nf_ro.properties_out[0].temperature,
        name="Solution temperature",
        ui_units=pyunits.K,
        display_units="K",
        rounding=2,
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    # Unit model data, NF Pump
    exports.add(
        obj=fs.dye_separation.P1.eta_motor,
        name="NF pump- motor efficiency",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="NF pump- motor efficiency",
        is_input=True,
        input_category="rHGO Nanofiltration",
        is_output=False,
    )
    exports.add(
        obj=fs.dye_separation.P1.eta_pump,
        name="NF pump efficiency",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="NF pump efficiency",
        is_input=True,
        input_category="rHGO Nanofiltration",
        is_output=False,
    )
    # Unit model data, rHGO Nanofiltration
    exports.add(
        obj=fs.dye_separation.nanofiltration.recovery_frac_mass_H2O[0],
        name="Water recovery",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Water recovery [g-H2O treated/g-H2O inlet]",
        is_input=True,
        input_category="rHGO Nanofiltration",
        is_output=False,
    )
    exports.add(
        obj=fs.dye_separation.nanofiltration.removal_frac_mass_comp[0, "dye"],
        name="Mass removal fraction, dye",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Dye mass removal [kg-Dye removed/kg-Dye inlet]",
        is_input=True,
        input_category="rHGO Nanofiltration",
        is_output=False,
    )
    exports.add(
        obj=fs.dye_separation.nanofiltration.removal_frac_mass_comp[0, "tds"],
        name="Mass removal fraction, TDS",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="TDS mass removal [kg-Dye removed/kg-Dye inlet]",
        is_input=True,
        input_category="rHGO Nanofiltration",
        is_output=False,
    )
    exports.add(
        obj=fs.dye_separation.nanofiltration.water_permeability_coefficient[0],
        name="NF Water permeability Coefficient, A",
        ui_units=pyunits.L / pyunits.m**2 / pyunits.hour / pyunits.bar,
        display_units="LMH/bar",
        rounding=2,
        description="Membrane water permeability Coefficient, A",
        is_input=True,
        input_category="rHGO Nanofiltration",
        is_output=False,
    )
    v = fs.dye_separation.nanofiltration.applied_pressure
    exports.add(
        obj=v[0],
        name=v.doc,
        ui_units=getattr(pyunits, str(v._units)),
        display_units=str(v._units),
        rounding=1,
        description=v.doc,
        is_input=True,
        input_category="rHGO Nanofiltration",
        is_output=False,
    )

    # Unit model, secondary WWTP
    if hasattr(fs, "pretreatment"):
        exports.add(
            obj=fs.pretreatment.wwtp.energy_electric_flow_vol_inlet,
            name="Specific energy consumption per inlet flow rate",
            ui_units=pyunits.kWh / pyunits.m**3,
            display_units="kWh/m3",
            rounding=2,
            description="Electrical energy consumption with respect to influent volumetric flow rate",
            is_input=True,
            input_category="Secondary Wastewater Treatment",
            is_output=False,
        )
    else:
        pass

    # Unit model, RO
    v = fs.desalination.RO.A_comp
    exports.add(
        obj=v[0, "H2O"],
        name="RO Water permeability coefficient, A",
        ui_units=getattr(pyunits, str(v._units)),
        display_units=str(v._units),
        rounding=12,
        description="Membrane water permeability Coefficient, A",
        is_input=True,
        input_category="Reverse Osmosis",
        is_output=False,
    )
    v = fs.desalination.RO.B_comp
    exports.add(
        obj=v[0, "TDS"],
        name="RO Salt permeability coefficient, B",
        ui_units=getattr(pyunits, str(v._units)),
        display_units=str(v._units),
        rounding=8,
        description="Membrane salt permeability Coefficient, B",
        is_input=True,
        input_category="Reverse Osmosis",
        is_output=False,
    )
    exports.add(
        obj=fs.desalination.P2.efficiency_pump[0],
        name="RO high-pressure pump efficiency",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="RO high-pressure pump efficiency",
        is_input=True,
        input_category="Reverse Osmosis",
        is_output=False,
    )
    exports.add(
        obj=fs.desalination.P3.efficiency_pump[0],
        name="RO booster pump efficiency",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="RO booster pump efficiency",
        is_input=True,
        input_category="Reverse Osmosis",
        is_output=False,
    )
    exports.add(
        obj=fs.desalination.PXR.efficiency_pressure_exchanger[0],
        name="Isobaric pressure exchanger efficiency",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Isobaric pressure exchanger efficiency",
        is_input=True,
        input_category="Reverse Osmosis",
        is_output=False,
    )
    # Unit cost data, NF pump
    exports.add(
        obj=fs.zo_costing.pump_electricity.pump_cost["default"],
        name="Pump cost",
        ui_units=fs.zo_costing.base_currency / (pyunits.m**3 / pyunits.hr),
        display_units="$/(m^3/hr)",
        rounding=0,
        description="Pump capital cost parameter",
        is_input=True,
        input_category="rHGO NF costing",
        is_output=False,
    )
    # Unit cost data, NF
    v = fs.zo_costing.nanofiltration.membrane_cost
    exports.add(
        obj=v["rHGO_dye_rejection"],
        name="NF membrane cost",
        ui_units=getattr(pyunits, str(v._units)),
        display_units=str(v._units),
        rounding=2,
        is_input=True,
        input_category="rHGO NF costing",
        is_output=False,
    )
    v = fs.zo_costing.nanofiltration.membrane_replacement_rate
    exports.add(
        obj=v["rHGO_dye_rejection"],
        name="NF membrane replacement rate",
        ui_units=getattr(pyunits, str(v._units)),
        display_units="fraction",
        rounding=2,
        is_input=True,
        input_category="rHGO NF costing",
        is_output=False,
    )
    # Unit cost data, RO
    v = fs.ro_costing.reverse_osmosis.membrane_cost
    exports.add(
        obj=v,
        name="RO membrane cost",
        ui_units=getattr(pyunits, str(v._units)),
        display_units=str(v._units),
        rounding=2,
        is_input=True,
        input_category="RO costing",
        is_output=False,
    )
    v = fs.ro_costing.reverse_osmosis.factor_membrane_replacement
    exports.add(
        obj=v,
        name="RO membrane replacement rate",
        ui_units=getattr(pyunits, str(v._units)),
        display_units="fraction",
        rounding=2,
        is_input=True,
        input_category="RO costing",
        is_output=False,
    )
    v = fs.ro_costing.high_pressure_pump.cost
    exports.add(
        obj=v,
        name="RO unit pump cost",
        ui_units=getattr(pyunits, str(v._units)),
        display_units=str(v._units),
        rounding=2,
        description="Unit pump cost for high-pressure and booster pump",
        is_input=True,
        input_category="RO costing",
        is_output=False,
    )
    v = fs.ro_costing.pressure_exchanger.cost
    exports.add(
        obj=v,
        name=v.doc,
        ui_units=getattr(pyunits, str(v._units)),
        display_units=str(v._units),
        rounding=2,
        is_input=True,
        input_category="RO costing",
        is_output=False,
    )
    # System costs
    v = fs.ro_costing.electricity_cost
    exports.add(
        obj=v,
        name="Electricity cost",
        ui_units=getattr(pyunits, str(v._units)),
        display_units=str(v._units),
        rounding=2,
        is_input=True,
        input_category="System Costs",
        is_output=False,
    )
    v = fs.zo_costing.waste_disposal_cost
    exports.add(
        obj=v,
        name="Waste disposal cost per volume",
        ui_units=getattr(pyunits, str(v._units)),
        display_units=str(v._units),
        rounding=2,
        is_input=True,
        input_category="System Costs",
        is_output=False,
    )
    v = fs.zo_costing.dye_disposal_cost
    exports.add(
        obj=v,
        name="Dye disposal cost per volume",
        ui_units=getattr(pyunits, str(v._units)),
        display_units=str(v._units),
        rounding=2,
        is_input=True,
        input_category="System Costs",
        is_output=False,
    )
    v = fs.zo_costing.recovered_water_cost
    exports.add(
        obj=v,
        name="Value of recovered water",
        ui_units=getattr(pyunits, str(v._units)),
        display_units=str(v._units),
        rounding=2,
        is_input=True,
        input_category="System Costs",
        is_output=False,
    )
    # NF Results for rejection and membrane area
    exports.add(
        obj=fs.dye_separation.nanofiltration.rejection_comp[0, "dye"],
        name="Solute Rejection- dye",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Rejection rate, dye",
        is_input=False,
        output_category="rHGO Nanofiltration",
        is_output=True,
    )
    exports.add(
        obj=fs.dye_separation.nanofiltration.rejection_comp[0, "tds"],
        name="Solute Rejection- tds",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Rejection rate, tds",
        is_input=False,
        output_category="rHGO Nanofiltration",
        is_output=True,
    )
    v = fs.dye_separation.nanofiltration.area
    exports.add(
        obj=fs.dye_separation.nanofiltration.area,
        name=v.doc,
        ui_units=getattr(pyunits, str(v._units)),
        display_units=str(v._units),
        rounding=2,
        description="NF Membrane area",
        is_input=False,
        output_category="rHGO Nanofiltration",
        is_output=True,
    )
    # Outlets
    exports.add(
        obj=fs.dye_retentate.properties[0].flow_vol,
        name="Volumetric NF retentate flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=2,
        description="NF retentate flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.dye_retentate.properties[0].conc_mass_comp["dye"],
        name="NF retentate concentration, dye",
        ui_units=pyunits.kg / pyunits.m**3,
        display_units="kg/m3",
        rounding=2,
        description="NF retentate concentration, total dissolved solids (TDS)",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.dye_retentate.properties[0].conc_mass_comp["tds"],
        name="NF retentate concentration, tds",
        ui_units=pyunits.kg / pyunits.m**3,
        display_units="kg/m3",
        rounding=2,
        description="NF retentate concentration, total dissolved solids (TDS)",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.permeate.properties[0].flow_vol,
        name="Volumetric RO permeate flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=2,
        description="RO permeate flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.permeate.properties[0].conc_mass_phase_comp["Liq", "TDS"],
        name="RO permeate concentration, TDS",
        ui_units=pyunits.kg / pyunits.m**3,
        display_units="kg/m3",
        rounding=2,
        description="RO permeate concentration, total dissolved solids (TDS), including residual dye,",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.brine.properties[0].flow_vol,
        name="Volumetric RO brine flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=2,
        description="RO brine flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.brine.properties[0].conc_mass_phase_comp["Liq", "TDS"],
        name="RO brine concentration, TDS",
        ui_units=pyunits.kg / pyunits.m**3,
        display_units="kg/m3",
        rounding=2,
        description="RO brine concentration, total dissolved solids (TDS), including residual dye,",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )

    # System metrics
    exports.add(
        obj=fs.LCOT_wo_revenue,
        name="Levelized cost of treatment",
        ui_units=fs.zo_costing.base_currency / pyunits.m**3,
        display_units="$/m3 of feed",
        rounding=2,
        description="Levelized cost of treatment with respect to influent flow",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )
    exports.add(
        obj=fs.LCOT,
        name="Levelized cost of treatment w/revenue and disposal costs",
        ui_units=fs.zo_costing.base_currency / pyunits.m**3,
        display_units="$/m3 of feed",
        rounding=2,
        description="Net levelized cost of treatment wrt to influent flow accounting for revenue and disposal costs",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )
    exports.add(
        obj=fs.LCOW_wo_revenue,
        name="Levelized cost of water",
        ui_units=fs.zo_costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=2,
        description="Levelized cost of water with respect to product water",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )
    exports.add(
        obj=fs.LCOW,
        name="Net levelized cost of water  w/revenue and disposal costs",
        ui_units=fs.zo_costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=2,
        description="Net levelized cost of water wrt to product flow accounting for revenue and disposal costs`",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )

    # Normalized metrics
    total_capital_norm = fs.total_capital_cost / fs.feed.properties[0].flow_vol
    exports.add(
        obj=total_capital_norm,
        name="Total capital",
        ui_units=fs.zo_costing.base_currency / (pyunits.m**3 / pyunits.day),
        display_units="$/(m3/day)",
        rounding=1,
        description="Normalized total capital costs accounting for indirect "
        "capital and installation - [total capital costs/feed flow rate]",
        is_input=False,
        is_output=True,
        output_category="Normalized cost metrics",
    )
    annual_water_inlet = (
        pyunits.convert(
            fs.feed.properties[0].flow_vol,
            to_units=pyunits.m**3 / pyunits.year,
        )
        * fs.zo_costing.utilization_factor
    )

    elec_operating_norm = (
        fs.zo_costing.aggregate_flow_costs["electricity"]
        + fs.ro_costing.aggregate_flow_costs["electricity"]
    ) / annual_water_inlet
    exports.add(
        obj=elec_operating_norm,
        name="Electricity",
        ui_units=fs.zo_costing.base_currency / pyunits.m**3,
        display_units="$/m3 of feed",
        rounding=2,
        description="Normalized electricity cost - [annual electricity costs/annual "
        "feed flow rate]",
        is_input=False,
        is_output=True,
        output_category="Normalized cost metrics",
    )

    # performance metrics
    recovery_vol = fs.permeate.properties[0].flow_vol / fs.feed.properties[0].flow_vol
    exports.add(
        obj=recovery_vol,
        name="Volumetric recovery of product water",
        ui_units=pyunits.dimensionless,
        display_units="m3 of product/m3 of feed",
        rounding=3,
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    dye_recovery = (
        fs.dye_retentate.flow_mass_comp[0, "dye"] / fs.feed.flow_mass_comp[0, "dye"]
    )
    exports.add(
        obj=dye_recovery,
        name="Dye mass recovery",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="Mass recovery of dye [mass dye in NF retentate/mass dye in feed]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )

    # Capital costs
    exports.add(
        obj=fs.total_capital_cost,
        name="Total",
        ui_units=fs.zo_costing.base_currency,
        display_units="$",
        rounding=2,
        description="Total capital costs - including investment factor to account "
        "for indirect capital",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )
    if hasattr(fs, "pretreatment"):
        wwtp_capex = fs.pretreatment.wwtp.costing.capital_cost
        exports.add(
            obj=wwtp_capex,
            name="Secondary WWTP Cost",
            ui_units=fs.zo_costing.base_currency,
            display_units="$",
            rounding=2,
            description="Secondary WWTP representative of conventional activated sludge with secondary clarifier",
            is_input=False,
            is_output=True,
            output_category="Capital costs",
        )
    else:
        pass
    nf_capex = (
        fs.dye_separation.nanofiltration.costing.capital_cost
        + fs.dye_separation.P1.costing.capital_cost
    )
    exports.add(
        obj=nf_capex,
        name="rHGO Nanofiltration system costs",
        ui_units=fs.zo_costing.base_currency,
        display_units="$",
        rounding=2,
        description="rHGO Nanofiltration system costs including pump",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )
    exports.add(
        obj=fs.ro_costing.total_capital_cost,
        name="RO system costs",
        ui_units=fs.zo_costing.base_currency,
        display_units="$",
        rounding=2,
        description="RO total investment cost, including pumps and ERD",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )

    # Operating costs
    exports.add(
        obj=fs.total_operating_cost,
        name="Total",
        ui_units=fs.zo_costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=2,
        description="Total annual operating costs - including electricity "
        "and fixed, excluding disposal costs",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )
    total_elec_cost = fs.zo_costing.aggregate_flow_costs["electricity"]
    +fs.ro_costing.aggregate_flow_costs["electricity"]
    exports.add(
        obj=total_elec_cost,
        name="Electricity",
        ui_units=fs.zo_costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=2,
        description="Annual electricity costs ",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )
    exports.add(
        obj=fs.brine_disposal_cost,
        name="RO brine disposal",
        ui_units=fs.zo_costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=2,
        description="Annual brine disposal",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )
    exports.add(
        obj=fs.dye_disposal_cost,
        name="Dye disposal",
        ui_units=fs.zo_costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=2,
        description="Annual dye disposal",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )
    if hasattr(fs, "pretreatment"):
        exports.add(
            obj=fs.sludge_disposal_cost,
            name="WWTP sludge disposal",
            ui_units=fs.zo_costing.base_currency / pyunits.year,
            display_units="$/year",
            rounding=2,
            description="Annual sludge disposal",
            is_input=False,
            is_output=True,
            output_category="Operating costs",
        )
    else:
        pass
    # Revenue
    exports.add(
        obj=fs.water_recovery_revenue,
        name="Water",
        ui_units=fs.zo_costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=2,
        description="Revenue from selling water or value of freshwater savings",
        is_input=False,
        is_output=True,
        output_category="Revenue",
    )


def build_flowsheet(build_options=None, **kwargs):
    # build and solve initial flowsheet
    m = build(include_pretreatment=False)

    set_operating_conditions(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    initialize_system(m)

    results = solve(m)
    assert_optimal_termination(results)

    add_costing(m)
    assert_degrees_of_freedom(m, 0)

    results = solve(m)
    assert_optimal_termination(results)
    return m


def solve_flowsheet(flowsheet=None):
    fs = flowsheet
    results = solve(fs)
    return results
